import os
import time
import torch
from mltu.preprocessors import ImageReader
from mltu.annotations.images import CVImage
from mltu.transformers import ImageResizer, ImageShowCV2, ImageNormalizer
from mltu.augmentors import RandomBrightness, RandomRotate, RandomErodeDilate, RandomSharpen, \
    RandomMirror, RandomFlip, RandomGaussianBlur, RandomSaltAndPepper, RandomDropBlock, RandomMosaic
from mltu.torch.model import Model
from mltu.torch.dataProvider import DataProvider
from mltu.torch.yolo.annotation import VOCAnnotationReader
from mltu.torch.yolo.preprocessors import YoloPreprocessor
from mltu.torch.yolo.loss import v8DetectionLoss
from mltu.torch.yolo.metrics import YoloMetrics
from mltu.torch.yolo.optimizer import build_optimizer, AccumulativeOptimizer
from mltu.torch.callbacks import EarlyStopping, ModelCheckpoint, ReduceLROnPlateau, TensorBoard, Model2onnx, WarmupCosineDecay

from ultralytics.nn.tasks import DetectionModel
from ultralytics.engine.model import Model as BaseModel


annotations_path = "Datasets/car-plate-detection/annotations"

dataset = [[None, os.path.join(annotations_path, f)] for f in os.listdir(annotations_path)]

# Make sure torch can see GPU device, it is not recommended to train with CPU
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

img_size = 416
labels = {0: "licence"}

# Create a data provider for the dataset
data_provider = DataProvider(
    dataset=dataset,
    skip_validation=True,
    batch_size=16,
    data_preprocessors=[
        VOCAnnotationReader(labels=labels),
        ImageReader(CVImage),
        ],
    transformers=[
        # ImageShowCV2(),
        ImageResizer(img_size, img_size),
        ImageNormalizer(transpose_axis=True),
    ],
    batch_postprocessors=[
        YoloPreprocessor(device, img_size)
    ],
    numpy=False,
)

# for b in data_provider:
#     pass

# split the dataset into train and test
train_data_provider, val_data_provider = data_provider.split(0.9, shuffle=False)

# Attaach augmentation to the train data provider
train_data_provider.augmentors = [
    RandomBrightness(), 
    RandomErodeDilate(),
    RandomSharpen(),
    RandomMirror(),
    RandomFlip(),
    RandomGaussianBlur(),
    RandomSaltAndPepper(),
    RandomRotate(angle=10),
    RandomDropBlock(),
    RandomMosaic(),
]

# for batch in train_data_provider:
#     pass
    # print(batch)
    # break



base_model = BaseModel("yolov8n.pt")
# Create a YOLO model
model = DetectionModel('yolov8n.yaml', nc=len(labels))

try: model.load_state_dict(base_model.model.state_dict(), strict=False)
except: pass

model.to(device)

for k, v in model.named_parameters():
    if any(x in k for x in [".dfl"]):
        print("freezing", k)
        v.requires_grad = False
    elif not v.requires_grad:
        v.requires_grad = True

lr = 1e-3
optimizer = build_optimizer(model.model, name="AdamW", lr=lr, weight_decay=0.0, momentum=0.937, decay=0.0005)
optimizer = AccumulativeOptimizer(optimizer, 16, 64)

# create model object that will handle training and testing of the network
model = Model(
    model, 
    optimizer, 
    v8DetectionLoss(model), 
    metrics=[YoloMetrics(nc=len(labels))],
    log_errors=False,
    output_path=f"Models/detector/{int(time.time())}",
    clip_grad_norm=10.0,
    ema=True,
)

modelCheckpoint = ModelCheckpoint(monitor="val_fitness", mode="max", save_best_only=True, verbose=True)
tensorBoard = TensorBoard()
earlyStopping = EarlyStopping(monitor="val_fitness", mode="max", patience=31, verbose=True)
model2onnx = Model2onnx(input_shape=(1, 3, img_size, img_size), verbose=True, opset_version=14, 
                    dynamic_axes = {"input": {0: "batch_size", 2: "height", 3: "width"}, 
                                    "output": {0: "batch_size", 2: "anchors"}},
                    metadata={"classes": labels})
warmupCosineDecayBias = WarmupCosineDecay(lr_after_warmup=lr, final_lr=lr, initial_lr=0.1, 
                                      warmup_steps=len(train_data_provider), warmup_epochs=10, ignore_param_groups=[1, 2]) # lr0
warmupCosineDecay = WarmupCosineDecay(lr_after_warmup=lr, final_lr=lr/10, initial_lr=1e-7, 
                                      warmup_steps=len(train_data_provider), warmup_epochs=10, decay_epochs=190, ignore_param_groups=[0]) # lr1 and lr2

# Train the model
history = model.fit(
    train_data_provider,
    test_dataProvider=val_data_provider,
    epochs=200, 
    callbacks=[
        modelCheckpoint, 
        tensorBoard, 
        earlyStopping, 
        model2onnx,
        warmupCosineDecayBias,
        warmupCosineDecay
        ]
    )


import tensorflow as tf
import numpy as np

from .attention import CrossAttention, GlobalSelfAttention, CausalSelfAttention

def positional_encoding(length: int, depth: int):
    """
    Generates a positional encoding for a given length and depth.

    Args:
        length (int): The length of the input sequence.
        depth (int): The depth that represents the dimensionality of the encoding.

    Returns:
        tf.Tensor: The positional encoding of shape (length, depth).
    """
    depth = depth / 2

    positions = np.arange(length)[:, np.newaxis]     # (seq, 1)
    depths = np.arange(depth)[np.newaxis, :]/depth   # (1, depth)

    angle_rates = 1 / (10000**depths)         # (1, depth)
    angle_rads = positions * angle_rates      # (pos, depth)

    pos_encoding = np.concatenate([np.sin(angle_rads), np.cos(angle_rads)], axis=-1) 

    return tf.cast(pos_encoding, dtype=tf.float32)


class PositionalEmbedding(tf.keras.layers.Layer):
    """
    A positional embedding layer combines the input embedding with a positional encoding that helps the Transformer
    to understand the relative position of the input tokens. This layer takes the input of tokens and converts them
    into sequence of embeddings vector. Then, it adds the positional encoding to the embeddings.

    Methods:
        compute_mask: Computes the mask to be applied to the embeddings.
        call: Performs the forward pass of the layer.
    """
    def __init__(self, vocab_size: int, d_model: int):
        """ Constructor of the PositionalEmbedding layer.

        Args:
            vocab_size (int): The size of the vocabulary. I. e. the number of unique tokens in the input sequence.
            d_model (int): The dimensionality of the embedding vector.
            embedding (tf.keras.layers.Embedding): The custom embedding layer. If None, a default embedding layer will be created.
        """
        super().__init__()
        self.d_model = d_model
        if vocab_size is not None:
            self.embedding = tf.keras.layers.Embedding(vocab_size, d_model, mask_zero=True)
        self.pos_encoding = positional_encoding(length=2048, depth=d_model)

    def compute_mask(self, *args, **kwargs):
        """ Computes the mask to be applied to the embeddings.

        Args:
            inputs: Tensor input(s).
            mask: Previous mask.

        Returns:
            tf.Tensor: The computed mask.
        """
        if hasattr(self, 'embedding'):
            return self.embedding.compute_mask(*args, **kwargs)
        else:
            return None  # If no embedding layer, return no mask

    def call(self, x: tf.Tensor) -> tf.Tensor:
        """ Performs the forward pass of the layer.
        
        Args:
            x (tf.Tensor): The input tensor of shape (batch_size, seq_length).

        Returns:
            tf.Tensor: The output sequence of embedding vectors with added positional information. The shape is
                (batch_size, seq_length, d_model).
        """
        if hasattr(self, 'embedding'):
            x = self.embedding(x)
        length = tf.shape(x)[1]
        # This factor sets the relative scale of the embedding and positonal_encoding.
        x *= tf.math.sqrt(tf.cast(self.d_model, x.dtype))
        x = x + tf.cast(self.pos_encoding[tf.newaxis, :length, :], dtype=x.dtype)
        return x
    
class FeedForward(tf.keras.layers.Layer):
    """
    A class that implements the feed-forward layer.

    Methods:
        call: Performs the forward pass of the layer.

    Attributes:
        seq (tf.keras.Sequential): The sequential layer that contains the feed-forward layers. It applies the two feed-forward layers and the dropout layer.
        add (tf.keras.layers.Add): The Add layer.
        layer_norm (tf.keras.layers.LayerNormalization): The LayerNormalization layer.
    """
    def __init__(self, d_model: int, dff: int, dropout_rate: float=0.1, activation: str='relu'):
        """
        Constructor of the FeedForward layer.

        Args:
            d_model (int): The dimensionality of the model.
            dff (int): The dimensionality of the feed-forward layer.
            dropout_rate (float): The dropout rate.
        """
        super().__init__()
        self.seq = tf.keras.Sequential([
            tf.keras.layers.Dense(dff, activation=activation),
            tf.keras.layers.Dense(d_model),
            tf.keras.layers.Dropout(dropout_rate)
        ])
        self.add = tf.keras.layers.Add()
        self.layer_norm = tf.keras.layers.LayerNormalization()

    def call(self, x: tf.Tensor) -> tf.Tensor:
        """
        The call function that performs the feed-forward operation. 

        Args:
            x (tf.Tensor): The input sequence of shape (batch_size, seq_length, d_model).

        Returns:
            tf.Tensor: The output sequence of shape (batch_size, seq_length, d_model).
        """
        x = self.add([x, self.seq(x)])
        x = self.layer_norm(x) 
        return x


class EncoderLayer(tf.keras.layers.Layer):
    """
    A single layer of the Encoder. Usually there are multiple layers stacked on top of each other.

    Methods:
        call: Performs the forward pass of the layer.

    Attributes:
        self_attention (GlobalSelfAttention): The global self-attention layer.
        ffn (FeedForward): The feed-forward layer.
    """
    def __init__(self, d_model: int, num_heads: int, dff: int, dropout_rate: float=0.1, activation: str='relu'):
        """
        Constructor of the EncoderLayer.

        Args:
            d_model (int): The dimensionality of the model.
            num_heads (int): The number of heads in the multi-head attention layer.
            dff (int): The dimensionality of the feed-forward layer.
            dropout_rate (float): The dropout rate.
        """
        super().__init__()

        self.self_attention = GlobalSelfAttention(
            num_heads=num_heads,
            key_dim=d_model,
            dropout=dropout_rate
            )

        self.ffn = FeedForward(d_model, dff, dropout_rate, activation)

    def call(self, x: tf.Tensor) -> tf.Tensor:
        """
        The call function that performs the forward pass of the layer.

        Args:
            x (tf.Tensor): The input sequence of shape (batch_size, seq_length, d_model).

        Returns:
            tf.Tensor: The output sequence of shape (batch_size, seq_length, d_model).
        """
        x = self.self_attention(x)
        x = self.ffn(x)
        return x

class Encoder(tf.keras.layers.Layer):
    """
    A custom TensorFlow layer that implements the Encoder. This layer is mostly used in the Transformer models 
    for natural language processing tasks, such as machine translation, text summarization or text classification.

    Methods:
        call: Performs the forward pass of the layer.

    Attributes:
        d_model (int): The dimensionality of the model.
        num_layers (int): The number of layers in the encoder.
        pos_embedding (PositionalEmbedding): The positional embedding layer.
        enc_layers (list): The list of encoder layers.
        dropout (tf.keras.layers.Dropout): The dropout layer.
    """
    def __init__(self, num_layers: int, d_model: int, num_heads: int, dff: int, vocab_size: int, dropout_rate: float=0.1, activation: str='relu', **kwargs):
        """
        Constructor of the Encoder.

        Args:
            num_layers (int): The number of layers in the encoder.
            d_model (int): The dimensionality of the model.
            num_heads (int): The number of heads in the multi-head attention layer.
            dff (int): The dimensionality of the feed-forward layer.
            vocab_size (int): The size of the vocabulary.
            dropout_rate (float): The dropout rate.
        """
        super().__init__(**kwargs)

        self.d_model = d_model
        self.num_layers = num_layers

        self.pos_embedding = PositionalEmbedding(vocab_size=vocab_size, d_model=d_model)

        self.enc_layers = [
            EncoderLayer(d_model=d_model,
                        num_heads=num_heads,
                        dff=dff,
                        dropout_rate=dropout_rate,
                        activation=activation)
            for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(dropout_rate)


    def call(self, x: tf.Tensor) -> tf.Tensor:
        """
        The call function that performs the forward pass of the layer.
        
        Args:
            x (tf.Tensor): The input sequence of shape (batch_size, seq_length).

        Returns:
            tf.Tensor: The output sequence of shape (batch_size, seq_length, d_model).
        """
        x = self.pos_embedding(x)  
        # here x has shape `(batch_size, seq_len, d_model)`

        # Add dropout.
        x = self.dropout(x)

        for i in range(self.num_layers):
            x = self.enc_layers[i](x)

        return x  # Shape `(batch_size, seq_len, d_model)`.


class DecoderLayer(tf.keras.layers.Layer):
    """
    A single layer of the Decoder. Usually there are multiple layers stacked on top of each other.
    
    Methods:
        call: Performs the forward pass of the layer.

    Attributes:
        causal_self_attention (CausalSelfAttention): The causal self-attention layer.
        cross_attention (CrossAttention): The cross-attention layer.
        ffn (FeedForward): The feed-forward layer.
    """
    def __init__(self, d_model: int, num_heads: int, dff: int, dropout_rate: float=0.1, activation: str='relu'):
        """
        Constructor of the DecoderLayer.

        Args:
            d_model (int): The dimensionality of the model.
            num_heads (int): The number of heads in the multi-head attention layer.
            dff (int): The dimensionality of the feed-forward layer.
            dropout_rate (float): The dropout rate. 
        """
        super(DecoderLayer, self).__init__()

        self.causal_self_attention = CausalSelfAttention(
            num_heads=num_heads,
            key_dim=d_model,
            dropout=dropout_rate)

        self.cross_attention = CrossAttention(
            num_heads=num_heads,
            key_dim=d_model,
            dropout=dropout_rate)

        self.ffn = FeedForward(d_model, dff, dropout_rate, activation=activation)

    def call(self, x: tf.Tensor, context: tf.Tensor) -> tf.Tensor:
        """
        The call function that performs the forward pass of the layer.

        Args:
            x (tf.Tensor): The input sequence of shape (batch_size, seq_length, d_model). x is usually the output of the previous decoder layer.
            context (tf.Tensor): The context sequence of shape (batch_size, seq_length, d_model). Context is usually the output of the encoder.
        """
        x = self.causal_self_attention(x=x)
        x = self.cross_attention(x=x, context=context)

        # Cache the last attention scores for plotting later
        self.last_attn_scores = self.cross_attention.last_attn_scores

        x = self.ffn(x)  # Shape `(batch_size, seq_len, d_model)`.
        return x

class Decoder(tf.keras.layers.Layer):
    """
    A custom TensorFlow layer that implements the Decoder. This layer is mostly used in the Transformer models
    for natural language processing tasks, such as machine translation, text summarization or text classification.

    Methods:
        call: Performs the forward pass of the layer.

    Attributes:
        d_model (int): The dimensionality of the model.
        num_layers (int): The number of layers in the decoder.
        pos_embedding (PositionalEmbedding): The positional embedding layer.
        dec_layers (list): The list of decoder layers.
        dropout (tf.keras.layers.Dropout): The dropout layer.
    """
    def __init__(self, num_layers: int, d_model: int, num_heads: int, dff: int, vocab_size: int, dropout_rate: float=0.1, activation: str='relu', **kwargs):
        """
        Constructor of the Decoder.

        Args:
            num_layers (int): The number of layers in the decoder.
            d_model (int): The dimensionality of the model.
            num_heads (int): The number of heads in the multi-head attention layer.
            dff (int): The dimensionality of the feed-forward layer.
            vocab_size (int): The size of the vocabulary.
            dropout_rate (float): The dropout rate.
        """
        super().__init__(**kwargs)

        self.d_model = d_model
        self.num_layers = num_layers

        self.pos_embedding = PositionalEmbedding(vocab_size=vocab_size, d_model=d_model)
        self.dropout = tf.keras.layers.Dropout(dropout_rate)
        self.dec_layers = [
            DecoderLayer(
                d_model=d_model, 
                num_heads=num_heads, 
                dff=dff, 
                dropout_rate=dropout_rate, 
                activation=activation) for _ in range(num_layers)]

        self.last_attn_scores = None

    def call(self, x: tf.Tensor, context: tf.Tensor) -> tf.Tensor:
        """
        The call function that performs the forward pass of the layer.

        Args:
            x (tf.Tensor): The input sequence of shape (batch_size, target_seq_len).
            context (tf.Tensor): The context sequence of shape (batch_size, input_seq_len, d_model).
        """
        # `x` is token-IDs shape (batch, target_seq_len)
        x = self.pos_embedding(x)  # (batch_size, target_seq_len, d_model)

        x = self.dropout(x)

        for i in range(self.num_layers):
            x  = self.dec_layers[i](x, context)

        self.last_attn_scores = self.dec_layers[-1].last_attn_scores

        # The shape of x is (batch_size, target_seq_len, d_model).
        return x

def Transformer(
    input_vocab_size: int, 
    target_vocab_size: int, 
    encoder_input_size: int = None,
    decoder_input_size: int = None,
    num_layers: int=6, 
    d_model: int=512, 
    num_heads: int=8,
    dff: int=2048,
    dropout_rate: float=0.1,
    ) -> tf.keras.Model:
    """
    A custom TensorFlow model that implements the Transformer architecture.

    Args:
        input_vocab_size (int): The size of the input vocabulary.
        target_vocab_size (int): The size of the target vocabulary.
        encoder_input_size (int): The size of the encoder input sequence.
        decoder_input_size (int): The size of the decoder input sequence.
        num_layers (int): The number of layers in the encoder and decoder.
        d_model (int): The dimensionality of the model.
        num_heads (int): The number of heads in the multi-head attention layer.
        dff (int): The dimensionality of the feed-forward layer.
        dropout_rate (float): The dropout rate.

    Returns:
        A TensorFlow Keras model.
    """
    encoder_input = tf.keras.layers.Input(shape=(encoder_input_size,), dtype=tf.int64)
    decoder_input = tf.keras.layers.Input(shape=(decoder_input_size,), dtype=tf.int64)

    encoder = Encoder(num_layers=num_layers, d_model=d_model, num_heads=num_heads, dff=dff, vocab_size=input_vocab_size, dropout_rate=dropout_rate)(encoder_input)
    decoder = Decoder(num_layers=num_layers, d_model=d_model, num_heads=num_heads, dff=dff, vocab_size=target_vocab_size, dropout_rate=dropout_rate)(decoder_input, encoder)

    output = tf.keras.layers.Dense(target_vocab_size)(decoder)

    return tf.keras.Model(inputs=[encoder_input, decoder_input], outputs=output)


import torch
from torch import nn
from ultralytics.nn.modules import C2f, Conv, Bottleneck

def infer_shortcut(bottleneck):
    c1 = bottleneck.cv1.conv.in_channels
    c2 = bottleneck.cv2.conv.out_channels
    return c1 == c2 and hasattr(bottleneck, 'add') and bottleneck.add

class C2f_v2(nn.Module):
    # CSP Bottleneck with 2 convolutions
    def __init__(self, c1, c2, n=1, shortcut=False, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        self.c = int(c2 * e)  # hidden channels
        self.cv0 = Conv(c1, self.c, 1, 1)
        self.cv1 = Conv(c1, self.c, 1, 1)
        self.cv2 = Conv((2 + n) * self.c, c2, 1)  # optional act=FReLU(c2)
        self.m = nn.ModuleList(Bottleneck(self.c, self.c, shortcut, g, k=((3, 3), (3, 3)), e=1.0) for _ in range(n))

    def forward(self, x):
        # y = list(self.cv1(x).chunk(2, 1))
        y = [self.cv0(x), self.cv1(x)]
        y.extend(m(y[-1]) for m in self.m)
        return self.cv2(torch.cat(y, 1))

def transfer_weights(c2f, c2f_v2):
    c2f_v2.cv2 = c2f.cv2
    c2f_v2.m = c2f.m

    state_dict = c2f.state_dict()
    state_dict_v2 = c2f_v2.state_dict()

    # Transfer cv1 weights from C2f to cv0 and cv1 in C2f_v2
    old_weight = state_dict['cv1.conv.weight']
    half_channels = old_weight.shape[0] // 2
    state_dict_v2['cv0.conv.weight'] = old_weight[:half_channels]
    state_dict_v2['cv1.conv.weight'] = old_weight[half_channels:]

    # Transfer cv1 batchnorm weights and buffers from C2f to cv0 and cv1 in C2f_v2
    for bn_key in ['weight', 'bias', 'running_mean', 'running_var']:
        old_bn = state_dict[f'cv1.bn.{bn_key}']
        state_dict_v2[f'cv0.bn.{bn_key}'] = old_bn[:half_channels]
        state_dict_v2[f'cv1.bn.{bn_key}'] = old_bn[half_channels:]

    # Transfer remaining weights and buffers
    for key in state_dict:
        if not key.startswith('cv1.'):
            state_dict_v2[key] = state_dict[key]

    # Transfer all non-method attributes
    for attr_name in dir(c2f):
        attr_value = getattr(c2f, attr_name)
        if not callable(attr_value) and '_' not in attr_name:
            setattr(c2f_v2, attr_name, attr_value)

    c2f_v2.load_state_dict(state_dict_v2)

def replace_c2f_with_c2f_v2(module):
    for name, child_module in module.named_children():
        if isinstance(child_module, C2f):
            # Replace C2f with C2f_v2 while preserving its parameters
            shortcut = infer_shortcut(child_module.m[0])
            c2f_v2 = C2f_v2(child_module.cv1.conv.in_channels, child_module.cv2.conv.out_channels,
                            n=len(child_module.m), shortcut=shortcut,
                            g=child_module.m[0].cv2.conv.groups,
                            e=child_module.c / child_module.cv2.conv.out_channels)
            transfer_weights(child_module, c2f_v2)
            setattr(module, name, c2f_v2)
        else:
            replace_c2f_with_c2f_v2(child_module)


import os
import numpy as np

class Audio:
    """ Audio object

    Attributes:
        audio (np.ndarray): Audio array
        sample_rate (int): Sample rate
        init_successful (bool): True if audio was successfully read
        library (object): Library used to read audio, tested only with librosa
    """
    init_successful = False
    augmented=False
    
    def __init__(
            self, 
            audioPath: str, 
            sample_rate: int=22050,
            library=None
        ) -> None:
        if library is None:
            raise ValueError("library must be provided. (e.g. librosa object)")

        if isinstance(audioPath, str):
            if not os.path.exists(audioPath):
                raise FileNotFoundError(f"Image {audioPath} not found.")

            self._audio, self.sample_rate = library.load(audioPath, sr=sample_rate)
            self.path = audioPath
            self.init_successful = True

        else:
            raise TypeError(f"audioPath must be path to audio file, not {type(audioPath)}")
        
    @property
    def audio(self) -> np.ndarray:
        return self._audio
    
    @audio.setter
    def audio(self, value: np.ndarray):
        self.augmented = True
        self._audio = value

    @property
    def shape(self) -> tuple:
        return self._audio.shape
    
    def numpy(self) -> np.ndarray:
        return self._audio
    
    def __add__(self, other: np.ndarray) -> np.ndarray:
        self._audio = self._audio + other
        self.augmented = True
        return self
    
    def __len__(self) -> int:
        return len(self._audio)
    
    def __call__(self) -> np.ndarray:
        return self._audio
    
    def __repr__(self):
        return repr(self._audio)
    
    def __array__(self):
        return self._audio


import tensorflow as tf

class BaseAttention(tf.keras.layers.Layer):
    """
    Base class for all attention layers. It contains the common functionality of all attention layers.
    This layer contains a MultiHeadAttention layer, a LayerNormalization layer and an Add layer.
    It is used as a base class for the GlobalSelfAttention, CausalSelfAttention and CrossAttention layers.
    And it is not intended to be used directly.

    Methods:
        call: Performs the forward pass of the layer.

    Attributes:
        mha (tf.keras.layers.MultiHeadAttention): The MultiHeadAttention layer.
        layernorm (tf.keras.layers.LayerNormalization): The LayerNormalization layer.
        add (tf.keras.layers.Add): The Add layer.
    """
    def __init__(self, **kwargs: dict):
        """ Constructor of the BaseAttention layer.
        
        Args:
            **kwargs: Additional keyword arguments that are passed to the MultiHeadAttention layer, e. g. 
                        num_heads (number of heads), key_dim (dimensionality of the key space), etc.
        """
        super().__init__()
        self.mha = tf.keras.layers.MultiHeadAttention(**kwargs)
        self.layernorm = tf.keras.layers.LayerNormalization()
        self.add = tf.keras.layers.Add()


class CrossAttention(BaseAttention):
    """
    A class that implements the cross-attention layer by inheriting from the BaseAttention class.
    This layer is used to process two different sequences and attends to the context sequence while processing the query sequence.

    Methods:
        call: Performs the forward pass of the layer.    

    Attributes:
        mha (tf.keras.layers.MultiHeadAttention): The MultiHeadAttention layer.
        layernorm (tf.keras.layers.LayerNormalization): The LayerNormalization layer.
        add (tf.keras.layers.Add): The Add layer.
    """
    def call(self, x: tf.Tensor, context: tf.Tensor) -> tf.Tensor:
        """
        The call function that performs the cross-attention operation.

        Args:
            x (tf.Tensor): The query (expected Transformer results) sequence of shape (batch_size, seq_length, d_model).
            context (tf.Tensor): The context (inputs to the Transformer) sequence of shape (batch_size, seq_length, d_model).

        Returns:
            tf.Tensor: The output sequence of shape (batch_size, seq_length, d_model).
        """
        attn_output, attn_scores = self.mha(query=x, key=context, value=context, return_attention_scores=True)

        # Cache the attention scores for plotting later.
        self.last_attn_scores = attn_scores

        x = self.add([x, attn_output])
        x = self.layernorm(x)

        return x


class GlobalSelfAttention(BaseAttention):
    """
    A class that implements the global self-attention layer by inheriting from the BaseAttention class.
    This layer is used to process a single sequence and attends to all the tokens in the sequence.

    Methods:
        call: Performs the forward pass of the layer.

    Attributes:
        mha (tf.keras.layers.MultiHeadAttention): The MultiHeadAttention layer.
        layernorm (tf.keras.layers.LayerNormalization): The LayerNormalization layer.
        add (tf.keras.layers.Add): The Add layer.
    """
    def call(self, x: tf.Tensor) -> tf.Tensor:
        """
        The call function that performs the global self-attention operation.

        Args:
            x (tf.Tensor): The input sequence of shape (batch_size, seq_length, d_model).

        Returns:
            tf.Tensor: The output sequence of shape (batch_size, seq_length, d_model).
        """
        attn_output = self.mha(query=x, value=x, key=x)
        x = self.add([x, attn_output])
        x = self.layernorm(x)
        return x


class CausalSelfAttention(BaseAttention):
    """
    Call self attention on the input sequence, ensuring that each position in the 
    output depends only on previous positions (i.e. a causal model).

    Methods:
        call: Performs the forward pass of the layer.

    Attributes:
        mha (tf.keras.layers.MultiHeadAttention): The MultiHeadAttention layer.
        layernorm (tf.keras.layers.LayerNormalization): The LayerNormalization layer.
        add (tf.keras.layers.Add): The Add layer.
    """
    def call(self, x: tf.Tensor) -> tf.Tensor:
        """
        The call function that performs the causal self-attention operation.
        
        Args:
            x (tf.Tensor): The input sequence of shape (batch_size, seq_length, d_model).

        Returns:
            tf.Tensor: The output sequence of shape (batch_size, seq_length, d_model).
        """
        attn_output = self.mha(query=x, value=x, key=x, use_causal_mask = True)
        x = self.add([x, attn_output])
        x = self.layernorm(x)
        return x


import os
import cv2
import time
import torch
import numpy as np
from tqdm import tqdm
from pathlib import Path
from ultralytics.nn.tasks import DetectionModel
# from ultralytics.data.dataset import YOLODataset
from mltu.torch.dataProvider import DataProvider
from mltu.preprocessors import ImageReader
from mltu.annotations.images import CVImage
from mltu.torch.model import Model
# model = YOLO('yolov8n.pt')
from metrics import YoloMetrics
from mltu.transformers import ImageResizer, ImageShowCV2
from mltu.annotations.detections import Detections, Detection
from mltu.augmentors import RandomBrightness, RandomRotate, RandomErodeDilate, RandomSharpen, \
    RandomMirror, RandomFlip, RandomGaussianBlur, RandomSaltAndPepper, RandomDropBlock, RandomMosaic
import torch.nn as nn

from mltu.torch.callbacks import EarlyStopping, ModelCheckpoint, ReduceLROnPlateau, TensorBoard, Model2onnx, WarmupCosineDecay
# from loss import v8DetectionLoss


train_path = "/home/rokbal/Personal/ultralytics/generated_data/train/images"
train_labels = "/home/rokbal/Personal/ultralytics/generated_data/train/labels"

val_path = "/home/rokbal/Personal/ultralytics/generated_data/valid/images"
val_labels = "/home/rokbal/Personal/ultralytics/generated_data/valid/labels"

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

dataset = []
for f in tqdm(os.listdir(train_path)):
    img_path = Path(os.path.join(train_path, f))
    txt_path = Path(os.path.join(train_labels, img_path.stem + ".txt"))
    if txt_path.exists():
        dataset.append([str(img_path), str(txt_path)])

val_dataset = []
for f in tqdm(os.listdir(val_path)):
    img_path = Path(os.path.join(val_path, f))
    txt_path = Path(os.path.join(val_labels, img_path.stem + ".txt"))
    if txt_path.exists():
        val_dataset.append([str(img_path), str(txt_path)])

def annotation_preprocessor(image, annotation):
    with open(annotation, "r") as f:
        data = f.readlines()

    dets = [Detection(
            bbox=[float(x) for x in line.split()[1:]],
            label="ID",
            confidence=1.0,
            relative=True
        ) for line in data]

    detections = Detections(
        labels={0: "ID"},
        width=image.width,
        height=image.height,
        image_path=image.path,
        detections=dets
    )

    return image, detections


def image_normalizer(image, annotation):
    img = image.numpy().transpose(2, 0, 1) / 255.0
    return img, annotation


class YoloPreprocessor:
    def __init__(self, imgsz=640):
        self.imgsz = imgsz

    def __call__(self, images, annotations):
        batch = {
            "ori_shape": [],
            "resized_shape": [],
            "cls": [],
            "bboxes": [],
            "batch_idx": [],
        }

        for i, (image, detections) in enumerate(zip(images, annotations)):
            batch["ori_shape"].append([detections.height, detections.width])
            batch["resized_shape"].append([self.imgsz, self.imgsz])
            for detection in detections:
                batch["cls"].append([detection.labelId])
                batch["bboxes"].append(detection.xywh)
                batch["batch_idx"].append(i)

        batch["cls"] = torch.tensor(batch["cls"]).to(device)
        batch["bboxes"] = torch.tensor(batch["bboxes"]).to(device)
        batch["batch_idx"] = torch.tensor(batch["batch_idx"]).to(device)

        return np.array(images), batch





img_size = 320
# Create a data provider for the dataset
train_data_provider = DataProvider(
    dataset=dataset,
    skip_validation=True,
    batch_size=16,
    data_preprocessors=[
        ImageReader(CVImage),
        annotation_preprocessor,
        ],
    transformers=[
        ImageShowCV2(),
        ImageResizer(img_size, img_size),
        image_normalizer,
    ],
    use_cache=True,
    batch_postprocessors=[
        YoloPreprocessor(img_size)
    ],
    use_multiprocessing=False,
    max_queue_size=1,
    workers=1,
    numpy=False,
    augmentors=[
        # RandomBrightness(), 
        # RandomErodeDilate(),
        # RandomSharpen(),
        # RandomMirror(0.5),
        # RandomFlip(0.5),
        # RandomGaussianBlur(),
        # RandomSaltAndPepper(),
        # RandomRotate(0.5, angle=10),
        # RandomDropBlock(0.5),
        RandomMosaic(random_chance=1.0)
    ]
)

for d in tqdm(train_data_provider):
    pass


# train_data_provider, val_data_provider = data_provider.split(0.8)

# Augment training data with random brightness, rotation and erode/dilate
train_data_provider.augmentors = [
    RandomBrightness(), 
    RandomErodeDilate(),
    RandomSharpen(),
    RandomMirror(0.5),
    RandomFlip(0.5),
    RandomGaussianBlur(),
    RandomSaltAndPepper(),

    # RandomRotate(angle=10), 
    # ImageShowCV2(),
]

val_data_provider = DataProvider(
    dataset=val_dataset,
    skip_validation=True,
    batch_size=16,
    data_preprocessors=[
        ImageReader(CVImage),
        annotation_preprocessor,
        ],
    transformers=[
        # ImageShowCV2(),
        ImageResizer(img_size, img_size),
        image_normalizer,
    ],
    use_cache=True,
    batch_postprocessors=[
        YoloPreprocessor(img_size)
    ],
    use_multiprocessing=False,
    max_queue_size=32,
    # workers=32,
    numpy=False,
)


# while True:
#     start = time.time()
#     for batch in tqdm(train_data_provider):
#         pass
#     end = time.time()
#     fps = len(data_provider) / (end - start)
#     print(f"FPS: {fps:.2f}")

# for i in range(1000):
#     pbar = tqdm(train_data_provider)
#     for d in pbar:
#         pass


# # model_y = YOLO('yolov8n.yaml')
# model = DetectionModel('yolov8n.yaml', nc=1) # YOLO('yolov8n.pt')
# model.to(device)
# model.load_state_dict(torch.load("Models/detector/1707665153/model.pt"))

# for k, v in model.named_parameters():
#     if any(x in k for x in [".dfl"]):
#         print("freezing", k)
#         v.requires_grad = False
#     elif not v.requires_grad:
#         v.requires_grad = True

# # model.model.nc = 1
# # optimizer = torch.optim.AdamW(model.parameters(), lr=1e-3, betas=(0.937, 0.999))

# pg0, pg1, pg2 = [], [], []  # optimizer parameter groups
# for k, v in model.named_modules():
#     if hasattr(v, 'bias') and isinstance(v.bias, nn.Parameter):
#         pg2.append(v.bias)  # biases
#     if isinstance(v, nn.BatchNorm2d):
#         pg0.append(v.weight)  # no decay
#     elif hasattr(v, 'weight') and isinstance(v.weight, nn.Parameter):
#         pg1.append(v.weight)  # apply decay

# optimizer = torch.optim.AdamW(pg0, lr=1e-2, betas=(0.937, 0.999))
# optimizer.add_param_group({'params': pg1, 'weight_decay': 0.0005})  # add pg1 with weight_decay
# optimizer.add_param_group({'params': pg2, 'weight_decay': 0.0})  # add pg2 (biases)
# # del pg0, pg1, pg2



# loss = v8DetectionLoss(model)
# # create model object that will handle training and testing of the network
# model = Model(
#     model, 
#     optimizer, 
#     loss, 
#     metrics=[YoloMetrics(nc=1)],
#     log_errors=False,
#     mixed_precision=False,
#     output_path=f"Models/detector/{int(time.time())}",
#     clip_grad_norm=10.0,
#     ema=True,
# )

# # ouput_path = f"Models/detector/{int(time.time())}"
# # save_model_path = ouput_path + "/model.pt"
# # modelCheckpoint = ModelCheckpoint(save_model_path, monitor="val_fitness", mode="max", save_best_only=True, verbose=True)
# # tensorBoard = TensorBoard(log_dir=ouput_path)
# # earlyStopping = EarlyStopping(monitor="val_fitness", mode="max", patience=21, verbose=True)
# # reduceLROnPlateau = ReduceLROnPlateau(monitor="val_fitness", factor=0.9, mode="max", patience=5, verbose=True)
# # model2onnx = Model2onnx(modelCheckpoint.filepath, input_shape=(1, 3, img_size, img_size), verbose=True, opset_version=14, dynamic_axes=None)
# # warmupCosineDecay = WarmupCosineDecay(lr_after_warmup=1e-3, final_lr=1e-3, warmup_epochs=5, decay_epochs=0, initial_lr=1e-5)

# # history = model.fit(
# #     train_data_provider,
# #     test_dataProvider=val_data_provider,
# #     epochs=1, 
# #     callbacks=[modelCheckpoint, tensorBoard, earlyStopping, model2onnx, warmupCosineDecay]
# #     )
# # for _ in data_provider:
# #     pass


# # data = {
# #     "train": train_path,
# # }

# # dataset = YOLODataset(
# #     img_path=train_path,
# #     imgsz=320,
# #     batch_size=4,
# # )


# import math
# from mltu.torch.yolo.tolov8_utils import replace_c2f_with_c2f_v2
# import torch_pruning as tp
# from ultralytics.nn.modules import Detect


# example_inputs = torch.randn(1, 3, img_size, img_size).to(model.device)
# macs_list, nparams_list, map_list, pruned_map_list = [], [], [], []
# replace_c2f_with_c2f_v2(model.model)
# model.model = model.model.to(model.device)
# model.model.train()
# base_macs, base_nparams = tp.utils.count_ops_and_params(model.model, example_inputs)
# macs_list.append(base_macs)

# # prune same ratio of filter based on initial size
# target_prune_rate = 0.5
# iterative_steps = 16
# pruning_ratio = 1 - math.pow((1 - target_prune_rate), 1 / iterative_steps)

# init_map = 0
# print(f"Before Pruning: MACs={base_macs / 1e9: .5f} G, #Params={base_nparams / 1e6: .5f} M, mAP={init_map: .5f}")

# ignored_layers = [m for m in model.model.modules() if isinstance(m, (Detect,))]
# output_path = model.output_path
# for i in range(iterative_steps):

#     model.model.train()
#     for name, param in model.model.named_parameters():
#         param.requires_grad = True


#     example_inputs = example_inputs.to(model.device)
#     pruner = tp.pruner.GroupNormPruner(
#         model.model,
#         example_inputs,
#         importance=tp.importance.GroupNormImportance(),  # L2 norm pruning,
#         iterative_steps=1,
#         pruning_ratio=pruning_ratio,
#         ignored_layers=ignored_layers,
#         unwrapped_parameters=[]
#     )

#     pruner.step()

#     pruned_macs, pruned_nparams = tp.utils.count_ops_and_params(pruner.model, example_inputs.to(model.device))
#     macs_list.append(pruned_macs)
#     current_speed_up = float(macs_list[0]) / pruned_macs
#     print(f"After pruning iter {i + 1}: MACs={pruned_macs / 1e9} G, #Params={pruned_nparams / 1e6} M, speed up={current_speed_up}")

#     # os.makedirs(ouput_path, exist_ok=True)
#     #save_model_path = ouput_path + f"/model_prune{i}.pth"
#     model.output_path = Path(os.path.join(output_path.parent, f"prune_{i}", output_path.name))
#     history = model.fit(
#         train_data_provider,
#         test_dataProvider=val_data_provider,
#         epochs=10, 
#         callbacks=[
#             TensorBoard(),
#             ModelCheckpoint(monitor="val_fitness", mode="max", save_best_only=True, verbose=True),
#             Model2onnx(input_shape=(1, 3, img_size, img_size), verbose=True, opset_version=14, dynamic_axes=None),
#             # WarmupCosineDecay(lr_after_warmup=1e-3, final_lr=1e-4, warmup_epochs=3, decay_epochs=7, initial_lr=1e-5)
#             #modelCheckpoint, tensorBoard, earlyStopping, model2onnx, warmupCosineDecay
#             ],
#         )
    
#     # reset model optimizer
#     # model.optimizer = torch.optim.AdamW(model.model.parameters(), lr=1e-3, betas=(0.937, 0.999))
#     # model.optimizer.add_param_group({'params': pg1, 'weight_decay': 0.0005})  # add pg1 with weight_decay
#     # model.optimizer.add_param_group({'params': pg2, 'weight_decay': 0.0})  # add pg2 (biases)

#     # torch.save(model.model, save_model_path)
#     model.model.to(model.device)
#     del pruner


import torch
import numpy as np
from mltu.torch.yolo.detectors.detector import BaseDetector
from mltu.annotations.detections import BboxType, Detection, Detections

class Detector(BaseDetector):
    def __init__(
            self, 
            model,
            input_width: int, 
            input_height: int, 
            classes: dict, 
            confidence_threshold: float=0.5, 
            iou_threshold: float=0.5, 
            device: str="cuda"
        ):
        super().__init__()
        self.model = model
        self.input_width = input_width
        self.input_height = input_height
        self.classes = classes
        self.confidence_threshold = confidence_threshold
        self.iou_threshold = iou_threshold
        self.device = torch.device(device if torch.cuda.is_available() else "cpu")
        self.model.to(self.device)
        self.model.eval()
        
        # Generate a color palette for the classes
        self.color_palette = np.random.uniform(0, 255, size=(len(self.classes), 3))

    def predict(self, image: np.ndarray, **kwargs) -> Detections:
        img_height, img_width, _ = image.shape

        # Preprocess the image
        preprocessed_image = self.preprocess(image, self.input_height, self.input_width)
        
        # Perform inference on the preprocessed image
        preds = self.model(torch.tensor(preprocessed_image).to(self.device))

        # Convert torch tensor to numpy array
        results = preds[0].cpu().detach().numpy()

        # Calculate the scaling factors for the bounding box coordinates
        x_factor, y_factor = img_width / self.input_width, img_height / self.input_height

        # Perform postprocessing on the predictions
        boxes, scores, class_ids = self.postprocess(results, x_factor, y_factor, self.confidence_threshold, self.iou_threshold)

        detections = []
        for bbox, conf, class_id in zip(boxes, scores, class_ids):
            detection = Detection(
                bbox = bbox,
                label = self.classes[class_id],
                labels = self.classes,
                bbox_type=BboxType.XYWH,
                confidence=conf,
                relative=False,
                width=img_width,
                height=img_height
            )
            detections.append(detection)

        return Detections(
            labels=self.classes,
            width=img_width, 
            height=img_height, 
            detections=detections, 
            color_palette=self.color_palette,
        )


import tensorflow as tf
from keras.callbacks import Callback
from mltu.tensorflow.callbacks import Model2onnx


class EncDecSplitCallback(Callback):
    """Callback to extract the encoder and decoder models from Transformer model and save them separately
    Also, this callback incorporates Model2onnx callback to convert the encoder and decoder models to ONNX format

    Args:
        model_path (str): Path to save the encoder and decoder models
        encoder_metadata (dict, optional): Metadata to save with the encoder model. Defaults to None.
        decoder_metadata (dict, optional): Metadata to save with the decoder model. Defaults to None.
    """

    def __init__(
        self,
        model_path: str,
        encoder_metadata: dict = None,
        decoder_metadata: dict = None,
        model_name = "model.h5"
    ):
        """Callback to extract the encoder and decoder models from Transformer model and save them separately"""
        super(EncDecSplitCallback, self).__init__()
        self.model_path = model_path
        self.encoder_metadata = encoder_metadata
        self.decoder_metadata = decoder_metadata
        self.model_name = model_name

    def on_train_end(self, epoch: int, logs: dict = None):
        try:
            # load best model weights
            self.model.load_weights(self.model_path + "/" + self.model_name)
            
            # extract encoder and decoder models
            encoder_model = tf.keras.Model(
                inputs=self.model.inputs[0], outputs=self.model.get_layer("encoder").output
            )
            decoder_model = tf.keras.Model(
                inputs=[self.model.inputs[1], self.model.get_layer("encoder").output],
                outputs=self.model.layers[-1].output,
            )

            # save encoder and decoder models
            encoder_model.save(self.model_path + "/encoder.h5")
            decoder_model.save(self.model_path + "/decoder.h5")

            # convert encoder and decoder models to onnx
            Model2onnx.model2onnx(encoder_model, self.model_path + "/encoder.onnx")
            Model2onnx.model2onnx(decoder_model, self.model_path + "/decoder.onnx")

            # save encoder and decoder metadata
            if self.encoder_metadata:
                Model2onnx.include_metadata(self.model_path + "/encoder.onnx", self.encoder_metadata)
            if self.decoder_metadata:
                Model2onnx.include_metadata(self.model_path + "/decoder.onnx", self.decoder_metadata)
        except Exception as e:
            print(e)
            pass



import torch
from torch import nn

from ultralytics.utils.loss import BboxLoss, xywh2xyxy
from ultralytics.utils.tal import TaskAlignedAssigner, dist2bbox, make_anchors

class v8DetectionLoss:
    """Criterion class for computing training losses."""

    def __init__(self, model, box: float=7.5, cls: float=0.5, dfl: float=1.5):  # model must be de-paralleled
        """Initializes v8DetectionLoss with the model, defining model-related properties and BCE loss function."""
        self.model = model
        device = next(model.parameters()).device  # get model device

        self.head = model.model[-1]  # Detect() module
        self.bce = nn.BCEWithLogitsLoss(reduction="none")
        self.stride = self.head.stride  # model strides
        self.nc = self.head.nc  # number of classes
        self.no = self.head.no
        self.reg_max = self.head.reg_max # max number of regression targets
        self.device = device

        self.use_dfl = self.head.reg_max > 1

        self.assigner = TaskAlignedAssigner(topk=10, num_classes=self.nc, alpha=0.5, beta=6.0)
        self.bbox_loss = BboxLoss(self.head.reg_max - 1, use_dfl=self.use_dfl).to(device)
        self.proj = torch.arange(self.head.reg_max, dtype=torch.float, device=device).to(device)

        self.box = box # box gain
        self.cls = cls # cls gain
        self.dfl = dfl # dfl gain

    def preprocess(self, targets, batch_size, scale_tensor):
        """Preprocesses the target counts and matches with the input batch size to output a tensor."""
        if targets.shape[0] == 0:
            out = torch.zeros(batch_size, 0, 5, device=self.device)
        else:
            i = targets[:, 0]  # image index
            _, counts = i.unique(return_counts=True)
            counts = counts.to(dtype=torch.int32)
            out = torch.zeros(batch_size, counts.max(), 5, device=self.device)
            for j in range(batch_size):
                matches = i == j
                n = matches.sum()
                if n:
                    out[j, :n] = targets[matches, 1:]
            out[..., 1:5] = xywh2xyxy(out[..., 1:5].mul_(scale_tensor))
        return out

    def bbox_decode(self, anchor_points, pred_dist):
        """Decode predicted object bounding box coordinates from anchor points and distribution."""
        if self.use_dfl:
            b, a, c = pred_dist.shape  # batch, anchors, channels
            self.proj = self.proj.to(pred_dist.device)
            pred_dist = pred_dist.view(b, a, 4, c // 4).softmax(3).matmul(self.proj.type(pred_dist.dtype))

        return dist2bbox(pred_dist, anchor_points, xywh=False)

    def __call__(self, preds, batch):
        """Calculate the sum of the loss for box, cls and dfl multiplied by batch size."""
        loss = torch.zeros(3, device=self.device)  # box, cls, dfl
        feats = preds[1] if isinstance(preds, tuple) else preds
        pred_distri, pred_scores = torch.cat([xi.view(feats[0].shape[0], self.no, -1) for xi in feats], 2).split(
            (self.reg_max * 4, self.nc), 1
        )

        pred_scores = pred_scores.permute(0, 2, 1).contiguous()
        pred_distri = pred_distri.permute(0, 2, 1).contiguous()

        dtype = pred_scores.dtype
        batch_size = pred_scores.shape[0]
        imgsz = torch.tensor(feats[0].shape[2:], device=self.device, dtype=dtype) * self.stride[0]  # image size (h,w)
        anchor_points, stride_tensor = make_anchors(feats, self.stride, 0.5)

        # Targets
        targets = torch.cat((batch["batch_idx"].view(-1, 1), batch["cls"].view(-1, 1), batch["bboxes"]), 1)
        targets = self.preprocess(targets.to(self.device), batch_size, scale_tensor=imgsz[[1, 0, 1, 0]])
        gt_labels, gt_bboxes = targets.split((1, 4), 2)  # cls, xyxy
        mask_gt = gt_bboxes.sum(2, keepdim=True).gt_(0)

        # Pboxes
        pred_bboxes = self.bbox_decode(anchor_points, pred_distri)  # xyxy, (b, h*w, 4)

        _, target_bboxes, target_scores, fg_mask, _ = self.assigner(
            pred_scores.detach().sigmoid(),
            (pred_bboxes.detach() * stride_tensor).type(gt_bboxes.dtype),
            anchor_points * stride_tensor,
            gt_labels,
            gt_bboxes,
            mask_gt,
        )

        target_scores_sum = max(target_scores.sum(), 1)

        # Cls loss
        loss[1] = self.bce(pred_scores, target_scores.to(dtype)).sum() / target_scores_sum  # BCE

        # Bbox loss
        if fg_mask.sum():
            target_bboxes /= stride_tensor
            loss[0], loss[2] = self.bbox_loss(
                pred_distri, pred_bboxes, anchor_points, target_bboxes, target_scores, target_scores_sum, fg_mask
            )

        loss[0] *= self.box  # box gain
        loss[1] *= self.cls  # cls gain
        loss[2] *= self.dfl  # dfl gain

        detailed_loss = {"box_loss": loss[0].detach(), "cls_loss": loss[1].detach(), "dfl_loss": loss[2].detach()}

        return loss.sum() * batch_size, detailed_loss  # loss(box, cls, dfl)


import numpy as np
from mltu.inferenceModel import OnnxInferenceModel
from mltu.torch.yolo.detectors.detector import BaseDetector
from mltu.annotations.detections import BboxType, Detection, Detections

class Detector(OnnxInferenceModel, BaseDetector):
    """ YOLOv8 detector using onnxruntime"""
    def __init__(
            self, 
            model_path: str,
            input_width: int, 
            input_height: int, 
            confidence_threshold: float=0.5, 
            iou_threshold: float=0.5, 
            classes: dict = None, 
            return_raw_output: bool=False,
            *args, **kwargs
        ):
        """
        Args:
            model_path (str): Path to the model file
            input_width (int): Input width to use for the model
            input_height (int): Input height to use for the model
            confidence_threshold (float, optional): Confidence threshold for filtering the predictions. Defaults to 0.5.
            iou_threshold (float, optional): Intersection over union threshold for filtering the predictions. Defaults to 0.5.
            classes (dict, optional): Dictionary of class names. Defaults to None.
            return_raw_output (bool, optional): Return raw output of the model (return bounding boxes, scores, and class ids). Defaults to False.
        """
        super().__init__(model_path, *args, **kwargs)
        self.input_width = input_width
        self.input_height = input_height
        self.confidence_threshold = confidence_threshold
        self.iou_threshold = iou_threshold
        self.return_raw_output = return_raw_output

        self.classes = classes or self.metadata.get("classes", None)
        if self.classes is None:
            raise ValueError("The classes must be provided")
        
        # Generate a color palette for the classes
        self.color_palette = np.random.uniform(0, 255, size=(len(self.classes), 3))

    def predict(self, image: np.ndarray, **kwargs) -> Detections:
        img_height, img_width, _ = image.shape

        # Preprocess the image
        preprocessed_image = self.preprocess(image, self.input_height, self.input_width)
        
        # Perform inference on the preprocessed image
        preds = self.model.run(self.output_names, {self.input_names[0]: preprocessed_image})

        # Convert torch tensor to numpy array
        results = preds[0][0]

        # Calculate the scaling factors for the bounding box coordinates
        x_factor, y_factor = img_width / self.input_width, img_height / self.input_height

        # Perform postprocessing on the predictions
        boxes, scores, class_ids = self.postprocess(results, x_factor, y_factor, self.confidence_threshold, self.iou_threshold)

        if self.return_raw_output:
            return boxes, scores, class_ids

        detections = []
        for bbox, conf, class_id in zip(boxes, scores, class_ids):
            detection = Detection(
                bbox = bbox,
                label = self.classes[class_id],
                labels = self.classes,
                bbox_type=BboxType.XYWH,
                confidence=conf,
                relative=False,
                width=img_width,
                height=img_height
            )
            detections.append(detection)

        return Detections(
            labels=self.classes,
            width=img_width, 
            height=img_height, 
            detections=detections, 
            color_palette=self.color_palette,
        )


import cv2
import typing
import numpy as np
from enum import Enum


class BboxType(Enum):
    XYWH = 1 # x center, y center, width, height
    XYXY = 2 # left, top, right, bottom
    LTWH = 3 # left, top, width, height


class Detection:
    """ Object to hold the information of a detection for simple manipulation and visualization
    """
    def __init__(
            self, 
            bbox, 
            label: str, 
            labels: dict={},
            bbox_type: BboxType=BboxType.XYWH,
            confidence: float=0.0, 
            image_path: str="", 
            width: int=None, 
            height: int=None,
            relative: bool=False
        ):
        """
        Args:
            bbox (list or np.ndarray): Bounding box coordinates
            label (str): Label of the detection
            labels (dict, optional): Dictionary of labels. Defaults to {}.
            bbox_type (BboxType, optional): Type of the bounding box coordinates. Defaults to BboxType.XYWH.
            confidence (float, optional): Confidence score of the detection. Defaults to 0.0.
            image_path (str, optional): Path to the image. Defaults to "".
            width (int, optional): Width of the image. Defaults to None.
            height (int, optional): Height of the image. Defaults to None.
            relative (bool, optional): Whether the bounding box coordinates are relative to the image size. Defaults to False.
        """
        self.bbox = np.array(bbox)
        self.label = label
        self.labels = labels
        self.bbox_type = bbox_type
        self.confidence = confidence
        self.image_path = image_path
        self.width = width
        self.height = height
        self.relative = relative

        self.augmented = False

        self._xywh = None
        self._xyxy = None

        self.validate()

    @property
    def labelId(self) -> int:
        return self.label2id(self.label)

    def label2id(self, label: str) -> int:
        labelId = {v: k for k, v in self.labels.items()}.get(label, None)
        if labelId is None:
            raise ValueError(f"label {label} not found in labels")
        
        return labelId
    
    @property
    def xywh(self):
        return self._xywh
    
    @xywh.setter
    def xywh(self, xywh: np.ndarray):
        if (xywh[:2] + xywh[2:] / 2 > 1).any():
            # fix the bbox to be in range [0, 1]
            self._xywh = self.xyxy2xywh(self.xywh2xyxy(xywh))
        else:
            self._xywh = xywh.clip(0, 1)

        self._xyxy = self.xywh2xyxy(self._xywh)

    @property
    def xyxy(self):
        return self._xyxy

    @staticmethod
    def xywh2xyxy(xywh: np.ndarray):
        """ Convert bounding box from x, y, width, height to x1, y1, x2, y2
        """
        x, y, w, h = xywh
        x, y = x - w / 2, y - h / 2
        return np.array([x, y, x+w, y+h]).clip(0, 1)
    
    @staticmethod
    def xyxy2xywh(xyxy: np.ndarray):
        """ Convert bounding box from x1, y1, x2, y2 to x, y, width, height
        """
        x, y, x2, y2 = xyxy
        w, h = x2 - x, y2 - y
        return np.array([x + w / 2, y + h / 2, w, h]).clip(0, 1)
    
    @staticmethod
    def ltwh2xywh(ltwh: np.ndarray):
        """ Convert bounding box from left, top, width, height to x, y, width, height
        """
        l, t, w, h = ltwh
        return np.array([l + w / 2, t + h / 2, w, h]).clip(0, 1)
    
    def validate(self):
        """ Validate the bounding box coordinates
        """
        assert self.bbox_type in BboxType, f"bbox_type must be one of {BboxType}"
        if not self.relative:
            if self.width is None or self.height is None:
                raise ValueError("width and height must be provided when relative is False")
            
            if (np.array(self.bbox) > 1.0).any():
                raise ValueError("bbox coordinates must be in range [0, 1] when relative is False")
            
            bbox = np.array(self.bbox) / np.array([self.width, self.height, self.width, self.height])

        else:
            bbox = self.bbox

        if self.bbox_type.name == "XYWH":
            self.xywh = bbox

        elif self.bbox_type.name == "XYXY":
            self.xywh = self.xyxy2xywh(bbox)

        elif self.bbox_type.name == "LTWH":
            self.xywh = self.ltwh2xywh(bbox)

        else:
            raise ValueError(f"bbox_type {self.bbox_type} not supported")
        
    def flip(self, direction: int):
        new_xywh = self.xywh
        if direction == 0: # mirror
            new_xywh[0] = 1 - new_xywh[0]

        elif direction == 1: # vertical
            new_xywh[1] = 1 - new_xywh[1]

        self.xywh = new_xywh
        
        self.augmented = True

    def dot(self, rotMat: np.ndarray, width: int, height: int):
        """ Apply transformation matrix to detection

        Args:
            matrix (np.ndarray): Transformation matrix
            width (int): Width of the image
            height (int): Height of the image

        Returns:
            Object with transformed coordinates
        """
        # get the four corners of the bounding box
        bb = np.array(self.xyxy) * np.array([self.width, self.height, self.width, self.height])
        bb = np.array(((bb[0],bb[1]),(bb[2],bb[1]),(bb[2],bb[3]),(bb[0],bb[3])))

        bb_rotated = np.vstack((bb.T, np.array((1,1,1,1)))) #Convert the array to [x,y,1] format to dot it with the rotMat
        bb_rotated = np.dot(rotMat, bb_rotated).T #Perform Dot product and get back the points in shape of (4,2)

        # get the new coordinates of the bounding box
        x_min = min(bb_rotated[:, 0])
        y_min = min(bb_rotated[:, 1])
        x_max = max(bb_rotated[:, 0])
        y_max = max(bb_rotated[:, 1])

        new_x = (x_min + x_max) / 2
        new_y = (y_min + y_max) / 2
        new_w = x_max - x_min
        new_h = y_max - y_min

        # Normalize to the new width and height
        new_x /= width
        new_y /= height
        new_w /= width
        new_h /= height

        self.xywh = np.array([new_x, new_y, new_w, new_h])

        self.width = width
        self.height = height
        self.augmented = True
        
        return self

    def applyToFrame(self, frame: np.ndarray, color: tuple=(0, 255, 0), thickness: int=2, **kwargs) -> np.ndarray:
        """ Draw the bounding box on the image
        """
        # Get the coordinates of the bounding box
        x, y, x2, y2 = (self.xyxy * np.array([self.width, self.height, self.width, self.height])).astype(np.int32)

        # Draw the bounding box on the image
        frame = cv2.rectangle(frame.copy(), (x, y), (x2, y2), color, thickness, **kwargs)

        label = f"{self.label}: {self.confidence:.2f}" if self.confidence > 0 else self.label

        # Calculate the dimensions of the label text
        (label_width, label_height), _ = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)

        label_y = y - 10 if y - 10 > label_height else y + 10

        # Draw a filled rectangle as the background for the label text
        cv2.rectangle(
            frame, (x, label_y - label_height), (x + label_width, label_y + label_height), color, cv2.FILLED
        )

        cv2.putText(frame, label, (x, label_y), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 0), 1, cv2.LINE_AA)

        return frame

    def json(self):
        return {
            "xywh": self.xywh.tolist(),
            "label": self.label,
            "confidence": self.confidence,
            "image_path": self.image_path,
            "width": self.width,
            "height": self.height
        }
    

class Detections:
    """ Object to hold the information of multiple detections for simple manipulation and visualization"""
    def __init__(
            self, 
            labels: dict,
            width: int,
            height: int,
            detections: typing.Iterable[Detection] = [],
            image_path: str = "",
            color_palette: list=[]
        ) -> None:
        """ Initialize the Detections object

        Args:
            labels (dict): Dictionary of labels
            width (int): Width of the image
            height (int): Height of the image
            detections (typing.Iterable[Detection], optional): List of detections. Defaults to [].
            image_path (str, optional): Path to the image. Defaults to "".
            color_palette (list, optional): List of colors to use for the bounding boxes. Defaults to [].
        """
        self.labels = labels
        self.width = width
        self.height = height
        self.detections = detections
        self.image_path = image_path
        self.color_palette = color_palette

        self.validate()

    def label2id(self, label: str) -> int:
        labelId = {v: k for k, v in self.labels.items()}.get(label, None)
        if labelId is None:
            raise ValueError(f"label {label} not found in labels")
        
        return labelId

    def validate(self):
        for detection in self.detections:
            if not isinstance(detection, Detection):
                raise TypeError(f"detections must be iterable of Detection, not {type(detection)}")
            
            detection.width = self.width
            detection.height = self.height
            detection.labels = self.labels
            detection.image_path = self.image_path

        if isinstance(self.labels, list):
            self.labels = {i: label for i, label in enumerate(self.labels)}

    def applyToFrame(self, image: np.ndarray, **kwargs: dict) -> np.ndarray:
        """ Draw the detections on the image """
        for detection in self.detections:
            color = self.color_palette[detection.labelId] if len(self.color_palette) == len(self.labels) else (0, 255, 0)
            image = detection.applyToFrame(image, color=color, **kwargs)
        
        return image

    def __iter__(self):
        return iter(self.detections)

    def __len__(self):
        return len(self.detections)

    def __getitem__(self, index: int):
        return self.detections[index]


import torch
from torch import nn

class AccumulativeOptimizer(torch.optim.Optimizer):
    def __init__(self, optimizer, batch_size, nbs=64):
        super(AccumulativeOptimizer, self).__init__(optimizer.param_groups, optimizer.defaults)
        self.optimizer = optimizer
        self.accumulation_steps = int(nbs / batch_size)
        self.current_step = 0

    def zero_grad(self):
        if self.current_step == 0:
            self.optimizer.zero_grad()

    def step(self):
        self.current_step += 1
        if self.current_step >= self.accumulation_steps:
            self.optimizer.step()
            self.current_step = 0
            self.optimizer.zero_grad()


def build_optimizer(model, name: str="AdamW", lr: float=1e-3, weight_decay: float=0.0, momentum: float=0.937, decay=0.0005):

    pg0, pg1, pg2 = [], [], []  # optimizer parameter groups
    bn = tuple(v for k, v in nn.__dict__.items() if "Norm" in k)  # normalization layers, i.e. BatchNorm2d()
    for module_name, module in model.named_modules():
        for param_name, param in module.named_parameters(recurse=False):
            fullname = f"{module_name}.{param_name}" if module_name else param_name
            if "bias" in fullname:  # bias (no decay)
                pg2.append(param)
            elif isinstance(module, bn):  # weight (no decay)
                pg1.append(param)
            else:  # weight (with decay)
                pg0.append(param)

    if name == "AdamW":
        optimizer = torch.optim.AdamW(pg2, lr=lr, weight_decay=weight_decay, betas=(momentum, 0.999))
    elif name == "Adam":
        optimizer = torch.optim.Adam(pg2, lr=lr, weight_decay=weight_decay, betas=(momentum, 0.999))
    elif name == "SGD":
        optimizer = torch.optim.SGD(pg2, lr=lr, weight_decay=weight_decay, momentum=0.9)
    else:
        raise ValueError(f"Optimizer {name} not supported!")
    
    optimizer.add_param_group({'params': pg0, 'weight_decay': decay})  # add pg1 with weight_decay
    optimizer.add_param_group({'params': pg1, 'weight_decay': 0.0})  # add pg2 (biases)

    del pg0, pg1, pg2

    return optimizer


import torch
import numpy as np

class YoloPreprocessor:
    def __init__(self, device, imgsz=640):
        self.device = device
        self.imgsz = imgsz

    def __call__(self, images, annotations):
        batch = {
            "ori_shape": [],
            "resized_shape": [],
            "cls": [],
            "bboxes": [],
            "batch_idx": [],
        }

        for i, (image, detections) in enumerate(zip(images, annotations)):
            batch["ori_shape"].append([detections.height, detections.width])
            batch["resized_shape"].append([self.imgsz, self.imgsz])
            for detection in detections:
                batch["cls"].append([detection.labelId])
                batch["bboxes"].append(detection.xywh)
                batch["batch_idx"].append(i)

        batch["cls"] = torch.tensor(batch["cls"]).to(self.device)
        batch["bboxes"] = torch.tensor(batch["bboxes"]).to(self.device)
        batch["batch_idx"] = torch.tensor(batch["batch_idx"]).to(self.device)

        return np.array(images), batch


from mltu.torch.metrics import Metric

import torch
import numpy as np
from ultralytics.utils import ops
from ultralytics.utils.metrics import ap_per_class
from ultralytics.utils.metrics import Metric as YoloMetric
from ultralytics.utils.metrics import ConfusionMatrix, box_iou


class YoloMetrics(Metric):
    """ Accuracy metric class
    
    Args:
        name (str, optional): name of metric. Defaults to 'YoloMetrics'.
    """
    def __init__(self, nc, name: str="YoloMetrics", conf:float=0.001, iou:float=0.7, max_det:int=300) -> None:
        super(YoloMetrics, self).__init__(name=name)
        self.nc = nc
        self.jdict = []
        self.confusion_matrix = ConfusionMatrix(nc=nc, conf=conf)
        self.stats = dict(tp=[], conf=[], pred_cls=[], target_cls=[])

        self.conf = conf
        self.iou = iou
        self.lb = []
        self.multi_label=True
        self.single_cls=False
        self.max_det = max_det
        self.seen = 0
        self.device = torch.device("cpu")
        self.iouv = torch.linspace(0.5, 0.95, 10)  # iou vector for mAP@0.5:0.95
        self.niou = self.iouv.numel()
        self.box = YoloMetric()
        self.loss_info = {"box_loss": [], "cls_loss": [], "dfl_loss": []}

    @property
    def keys(self):
        return ["precision", "recall", "mAP50", "mAP50-95", "fitness"]

    def reset(self):
        self.__init__(self.nc, self.name)

    def match_predictions(self, pred_classes, true_classes, iou, use_scipy=False):
        """
        Matches predictions to ground truth objects (pred_classes, true_classes) using IoU.

        Args:
            pred_classes (torch.Tensor): Predicted class indices of shape(N,).
            true_classes (torch.Tensor): Target class indices of shape(M,).
            iou (torch.Tensor): An NxM tensor containing the pairwise IoU values for predictions and ground of truth
            use_scipy (bool): Whether to use scipy for matching (more precise).

        Returns:
            (torch.Tensor): Correct tensor of shape(N,10) for 10 IoU thresholds.
        """
        # Dx10 matrix, where D - detections, 10 - IoU thresholds
        correct = np.zeros((pred_classes.shape[0], self.iouv.shape[0])).astype(bool)
        # LxD matrix where L - labels (rows), D - detections (columns)
        correct_class = true_classes[:, None] == pred_classes
        iou = iou * correct_class  # zero out the wrong classes
        iou = iou.cpu().numpy()
        for i, threshold in enumerate(self.iouv.cpu().tolist()):
            if use_scipy:
                # WARNING: known issue that reduces mAP in https://github.com/ultralytics/ultralytics/pull/4708
                import scipy  # scope import to avoid importing for all commands

                cost_matrix = iou * (iou >= threshold)
                if cost_matrix.any():
                    labels_idx, detections_idx = scipy.optimize.linear_sum_assignment(cost_matrix, maximize=True)
                    valid = cost_matrix[labels_idx, detections_idx] > 0
                    if valid.any():
                        correct[detections_idx[valid], i] = True
            else:
                matches = np.nonzero(iou >= threshold)  # IoU > threshold and classes match
                matches = np.array(matches).T
                if matches.shape[0]:
                    if matches.shape[0] > 1:
                        matches = matches[iou[matches[:, 0], matches[:, 1]].argsort()[::-1]]
                        matches = matches[np.unique(matches[:, 1], return_index=True)[1]]
                        matches = matches[np.unique(matches[:, 0], return_index=True)[1]]
                    correct[matches[:, 1].astype(int), i] = True
        return torch.tensor(correct, dtype=torch.bool, device=pred_classes.device)

    def _process_batch(self, detections, gt_bboxes, gt_cls):
        """
        Return correct prediction matrix.

        Args:
            detections (torch.Tensor): Tensor of shape [N, 6] representing detections.
                Each detection is of the format: x1, y1, x2, y2, conf, class.
            labels (torch.Tensor): Tensor of shape [M, 5] representing labels.
                Each label is of the format: class, x1, y1, x2, y2.

        Returns:
            (torch.Tensor): Correct prediction matrix of shape [N, 10] for 10 IoU levels.
        """
        iou = box_iou(gt_bboxes, detections[:, :4])
        return self.match_predictions(detections[:, 5], gt_cls, iou)

    def _prepare_batch(self, si, batch):
        """Prepares a batch of images and annotations for validation."""
        idx = batch["batch_idx"] == si
        cls = batch["cls"][idx].squeeze(-1).to(self.device)
        bbox = batch["bboxes"][idx].to(self.device)
        ori_shape = batch["ori_shape"][si]
        # imgsz = batch["img"].shape[2:]
        imgsz = batch["resized_shape"][si]
        ratio_pad = (batch["resized_shape"][si][0] / batch["ori_shape"][si][0], batch["resized_shape"][si][1] / batch["ori_shape"][si][1])
        if len(cls):
            bbox = ops.xywh2xyxy(bbox) * torch.tensor(imgsz, device=self.device)[[1, 0, 1, 0]]  # target boxes
        return dict(cls=cls, bbox=bbox, ori_shape=ori_shape, imgsz=imgsz, ratio_pad=ratio_pad)

    def update_stats(self, preds, batch):
        """Metrics."""
        for si, pred in enumerate(preds):
            self.seen += 1
            npr = len(pred)
            stat = dict(
                conf=torch.zeros(0, device=self.device),
                pred_cls=torch.zeros(0, device=self.device),
                tp=torch.zeros(npr, self.niou, dtype=torch.bool, device=self.device),
            )
            pbatch = self._prepare_batch(si, batch)
            cls, bbox = pbatch.pop("cls"), pbatch.pop("bbox")
            nl = len(cls)
            stat["target_cls"] = cls
            if npr == 0:
                if nl:
                    for k in self.stats.keys():
                        self.stats[k].append(stat[k])
                continue

            # Predictions
            if self.single_cls:
                pred[:, 5] = 0
            predn = pred.clone().to(self.device)
            stat["conf"] = predn[:, 4]
            stat["pred_cls"] = predn[:, 5]

            # Evaluate
            if nl:
                stat["tp"] = self._process_batch(predn, bbox, cls)
            for k in self.stats.keys():
                self.stats[k].append(stat[k])

    def postprocess(self, preds):
        """Apply Non-maximum suppression to prediction outputs."""
        return ops.non_max_suppression(
            preds,
            self.conf,
            self.iou,
            labels=self.lb,
            multi_label=self.multi_label,
            agnostic=self.single_cls,
            max_det=self.max_det,
        )

    def update(self, output: torch.Tensor, target: torch.Tensor, **kwargs):
        """ Update metric state with new data

        Args:
            output (torch.Tensor): output of model
            target (torch.Tensor): target of data
        """
        # outputs = np.squeeze(output)
        model = kwargs.get("model")
        loss_info = kwargs.get("loss_info", None)
        if loss_info:
            for k, v in loss_info[0].items():
                self.loss_info[k].append(v.cpu().numpy())

        if not model.training: # update metrics only in validation mode
            preds = self.postprocess(output[0])
            self.update_stats(preds, target)

    def result(self) -> dict:
        """ Return metric value"""
        stats = {k: torch.cat(v, 0).cpu().numpy() for k, v in self.stats.items() if v}
        loss_info_stats = {k: np.mean(v) for k, v in self.loss_info.items()}
        if not stats:
            if loss_info_stats:
                return loss_info_stats
            return None
        
        # Compute statistics
        ap_per_class_results = ap_per_class(
            stats["tp"],
            stats["conf"],
            stats["pred_cls"],
            stats["target_cls"],
            plot=False,
            names={},
        )[2:]
        self.box.nc = self.nc
        self.box.update(ap_per_class_results)

        results = dict(zip(self.keys, self.box.mean_results() + [self.box.fitness()]))
        if loss_info_stats:
            results.update(loss_info_stats)

        return results


import os
import math
import torch
import typing
import traceback
import numpy as np
from torch import nn
from copy import deepcopy
from qqdm import qqdm, format_str
from pathlib import Path

from .metrics import Metric
from .callbacks import Callback
from .dataProvider import DataProvider
from .handlers import MetricsHandler, CallbacksHandler

def toTorch(data: np.ndarray, target: np.ndarray) -> typing.Tuple[torch.Tensor, torch.Tensor]:
    """ Check if data is of type torch.Tensor, if not convert it to torch.Tensor

    Args:
        data (np.ndarray): data to be converted
        target (np.ndarray): target to be converted

    Returns:
        typing.Tuple[torch.Tensor, torch.Tensor]: converted data and target
    """
    if not isinstance(data, torch.Tensor):
        data = torch.from_numpy(data)

    if not isinstance(target, (torch.Tensor, dict)):
        target = torch.from_numpy(target)

    if data.dtype != torch.float32:
        data = data.float()

    return data, target


def is_parallel(model):
    """Returns True if model is of type DP or DDP."""
    return isinstance(model, (nn.parallel.DataParallel, nn.parallel.DistributedDataParallel))

def de_parallel(model):
    """De-parallelize a model: returns single-GPU model if model is of type DP or DDP."""
    return model.module if is_parallel(model) else model

def copy_attr(a, b, include=(), exclude=()):
    """Copies attributes from object 'b' to object 'a', with options to include/exclude certain attributes."""
    for k, v in b.__dict__.items():
        if (len(include) and k not in include) or k.startswith("_") or k in exclude:
            continue
        else:
            setattr(a, k, v)

class ModelEMA:
    """Updated Exponential Moving Average (EMA) from https://github.com/rwightman/pytorch-image-models
    Keeps a moving average of everything in the model state_dict (parameters and buffers)
    For EMA details see https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage
    To disable EMA set the `enabled` attribute to `False`.
    """

    def __init__(self, model, decay=0.9999, tau=2000, updates=0):
        """Create EMA."""
        self.ema = deepcopy(de_parallel(model)).eval()  # FP32 EMA
        self.updates = updates  # number of EMA updates
        self.decay = lambda x: decay * (1 - math.exp(-x / tau))  # decay exponential ramp (to help early epochs)
        for p in self.ema.parameters():
            p.requires_grad_(False)
        self.enabled = True

    def update(self, model):
        """Update EMA parameters."""
        if self.enabled:
            self.updates += 1
            d = self.decay(self.updates)

            msd = de_parallel(model).state_dict()  # model state_dict
            for k, v in self.ema.state_dict().items():
                if v.dtype.is_floating_point:  # true for FP16 and FP32
                    v *= d
                    v += (1 - d) * msd[k].detach()
                    # assert v.dtype == msd[k].dtype == torch.float32, f'{k}: EMA {v.dtype},  model {msd[k].dtype}'

    def update_attr(self, model, include=(), exclude=("process_group", "reducer")):
        """Updates attributes and saves stripped model with optimizer removed."""
        if self.enabled:
            copy_attr(self.ema, model, include, exclude)


class Model:
    """ Model class for training and testing PyTorch neural networks"""
    def __init__(
        self, 
        model: torch.nn.Module, 
        optimizer: torch.optim.Optimizer, 
        loss: typing.Callable,
        metrics: typing.List[Metric] = [],
        mixed_precision: bool = False,
        scaler = None,
        log_errors: bool = True,
        output_path: str = None,
        clip_grad_norm: float = None,
        ema = False,
        ):
        """ Initialize model class

        Attributes:
            model (torch.nn.Module): PyTorch neural network
            optimizer (torch.optim.Optimizer): PyTorch optimizer
            loss (typing.Callable): loss function
            metrics (typing.List[Metric], optional): list of metrics. Defaults to [].
            mixed_precision (bool, optional): whether to use mixed precision. Defaults to False.
            scaler (torch.cuda.amp.GradScaler, optional): PyTorch GradScaler. Defaults to None.
            log_errors (bool, optional): whether to log errors. Defaults to True.
            output_path (str, optional): path to save model. Defaults to None.
            clip_grad_norm (float, optional): gradient clipping value. Defaults to None.
            ema (bool, optional): whether to use Exponential Moving Average. Defaults to False.
        """
        self.model = model
        self.optimizer = optimizer
        self.loss = loss
        # get device on which model is running
        self._device = next(self.model.parameters()).device

        self.metrics = MetricsHandler(metrics)

        self.mixed_precision = mixed_precision
        self.scaler = torch.cuda.amp.GradScaler() if mixed_precision else scaler
        self.log_errors = log_errors

        self.output_path = Path(output_path) if output_path else None
        if self.output_path:
            if self.output_path.suffix == "":
                self.output_path = Path(os.path.join(self.output_path, "model.pt"))
            os.makedirs(self.output_path.parent, exist_ok=True)

        self.stop_training = False
        self.clip_grad_norm = clip_grad_norm
        self.ema = ModelEMA(self.model) if ema else None

        self.validate()

    @property
    def device(self):
        return self._device

    def validate(self):
        """ Validate model, optimizer"""
        if not isinstance(self.model, torch.nn.Module):
            raise TypeError("model argument must be a torch.nn.Module")
        
        if not isinstance(self.optimizer, torch.optim.Optimizer):
            raise TypeError("optimizer argument must be a torch.optim.Optimizer")
        
    def toDevice(self, data: np.ndarray, target: np.ndarray) -> typing.Tuple[torch.Tensor, torch.Tensor]:
        """ Check if data is on the same device as model, if not move it to the device

        Args:
            data (np.ndarray): data to be moved
            target (np.ndarray): target to be moved

        Returns:
            typing.Tuple[torch.Tensor, torch.Tensor]: moved data and target
        """
        try:
            if data.device != self._device:
                data = data.to(self._device)

            if target.device != self._device:
                target = target.to(self._device)
        except AttributeError:
            if self.log_errors:
                print("Data and target must be of type torch.Tensor")

        return data, target

    def train_step(
        self, 
        data: typing.Union[np.ndarray, torch.Tensor], 
        target: typing.Union[np.ndarray, torch.Tensor],
        loss_info: dict = {}
        ) -> torch.Tensor:
        """ Perform one training step

        Args:
            data (typing.Union[np.ndarray, torch.Tensor]): training data
            target (typing.Union[np.ndarray, torch.Tensor]): training target
            loss_info (dict, optional): additional loss information. Defaults to {}.

        Returns:
            torch.Tensor: loss
        """
        self.optimizer.zero_grad()
        if self.mixed_precision:
            with torch.cuda.amp.autocast():
                output = self.model(data)
                loss = self.loss(output, target)
                if isinstance(loss, tuple):
                    loss, loss_info = loss[0], loss[1:]
                self.scaler.scale(loss).backward()
                if self.clip_grad_norm:
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=self.clip_grad_norm)
                self.scaler.step(self.optimizer)
                self.scaler.update()

        else:
            output = self.model(data)
            loss = self.loss(output, target)
            if isinstance(loss, tuple):
                loss, loss_info = loss[0], loss[1:]
            loss.backward()
            if self.clip_grad_norm:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=self.clip_grad_norm)
            self.optimizer.step()

        if self.ema:
            try:
                self.ema.update(self.model)
            except RuntimeError:
                self.ema = ModelEMA(self.model)
                self.ema.update(self.model)

        if self._device.type == "cuda":
            torch.cuda.synchronize() # synchronize after each forward and backward pass

        self.metrics.update(target, output, model=self.model, loss_info=loss_info)

        return loss
    
    def test_step(
        self, 
        data: typing.Union[np.ndarray, torch.Tensor], 
        target: typing.Union[np.ndarray, torch.Tensor],
        loss_info: dict = {}
        ) -> torch.Tensor:
        """ Perform one validation step

        Args:
            data (typing.Union[np.ndarray, torch.Tensor]): validation data
            target (typing.Union[np.ndarray, torch.Tensor]): validation target
            loss_info (dict, optional): additional loss information. Defaults to {}.

        Returns:
            torch.Tensor: loss
        """
        output = self.ema.ema(data) if self.ema else self.model(data)
        loss = self.loss(output, target)
        if isinstance(loss, tuple):
            loss, loss_info = loss[0], loss[1:]

        self.metrics.update(target, output, model=self.model, loss_info=loss_info)

        # clear GPU memory cache after each validation step
        torch.cuda.empty_cache()

        return loss
    
    def train(self, dataProvider: DataProvider):
        """ Perform one training epoch
        
        Args:
            dataProvider (DataProvider): data provider for training data

        Returns:
            dict: training results
        """
        # set model to training mode
        self.model.train()

        loss_sum = 0
        pbar = qqdm(dataProvider, total=len(dataProvider), desc=format_str('bold', f"Epoch {self._epoch}: "))
        for step, (data, target) in enumerate(pbar, start=1):
            self.callbacks.on_batch_begin(step, logs=None, train=True)

            data, target = self.toDevice(*toTorch(data, target))
            loss = self.train_step(data, target)
            loss_sum += loss.item()
            loss_mean = loss_sum / step

            # get training results of one step
            logs = self.metrics.results(loss_mean, train=True)

            # log learning rate into logs
            if len(self.optimizer.param_groups) > 1:
                lr_logs = {f"lr{i}": round(group["lr"], 6) for i, group in enumerate(self.optimizer.param_groups)}
                logs.update(lr_logs)
            else:
                logs["lr"] = round(self.optimizer.param_groups[0]["lr"], 6)

            # update progress bar description
            pbar.set_description(desc=format_str('bold', f"Epoch {self._epoch}: "))
            pbar.set_infos(logs)

            self.callbacks.on_batch_end(step, logs=logs, train=True)

        # reset metrics after each training epoch
        self.metrics.reset()

        # call on_epoch_end of data provider
        dataProvider.on_epoch_end()

        return logs

    def test(self, dataProvider: DataProvider):
        """ Perform one validation epoch

        Args:
            dataProvider (DataProvider): data provider for validation data

        Returns:
            dict: validation results
        """
        # set model to evaluation mode
        self.model.eval()
        loss_sum = 0
        pbar = qqdm(dataProvider, total=len(dataProvider), desc=format_str('bold', 'Description'))
        # disable autograd and gradient computation in PyTorch
        with torch.no_grad():
            for step, (data, target) in enumerate(pbar, start=1):
                self.callbacks.on_batch_begin(step, logs=None, train=False)

                data, target = self.toDevice(*toTorch(data, target))
                loss = self.test_step(data, target)
                loss_sum += loss.item()
                loss_mean = loss_sum / step

                # get testing results of one step
                logs = self.metrics.results(loss_mean, train=False)

                # update progress bar description
                pbar.set_description(f"Epoch {self._epoch}: ")
                pbar.set_infos(logs)

                self.callbacks.on_batch_end(step, logs=logs, train=False)

        # reset metrics after each test epoch
        self.metrics.reset()

        # call on_epoch_end of data provider
        dataProvider.on_epoch_end()

        return logs
    
    def save(self, path: str=None):
        """ Save model state dict to file

        Args:
            path (str): path to file
        """
        if not path and not self.output_path:
            print("Path to file is not provided, model will not be saved") # replace to error logging
            return
        
        model_to_save = self.ema.ema if self.ema else self.model
        
        output_path = Path(path or self.output_path)
        os.makedirs(output_path.parent, exist_ok=True)
        model_to_save.eval()
        try:
            torch.save(model_to_save.state_dict(), output_path)
            return str(output_path)
        except Exception:
            traceback.print_exc()
            torch.save(model_to_save, output_path.with_suffix(".pth"))
            return str(output_path.with_suffix(".pth"))
    
    def fit(
        self, 
        train_dataProvider: DataProvider, 
        test_dataProvider: DataProvider=None, 
        epochs: int=10, 
        initial_epoch: int = 1, 
        callbacks: typing.List[Callback] = []
        ) -> dict:
        """ Train model for a given number of epochs
        
        Args:
            train_dataProvider (DataProvider): data provider for training data
            test_dataProvider (DataProvider): data provider for validation data
            epochs (int): number of epochs
            initial_epoch (int, optional): initial epoch. Defaults to 1.
            callbacks (typing.List[Callback], optional): list of callbacks. Defaults to [].

        Returns:
            dict: training results
        """
        self._epoch = initial_epoch
        history = {}
        self.callbacks = CallbacksHandler(self, callbacks)
        self.callbacks.on_train_begin()
        for epoch in range(initial_epoch, initial_epoch + epochs):
            self.callbacks.on_epoch_begin(epoch)

            train_logs = self.train(train_dataProvider)
            val_logs = self.test(test_dataProvider) if test_dataProvider else {}

            logs = {**train_logs, **val_logs}
            self.callbacks.on_epoch_end(epoch, logs=logs)

            if self.stop_training:
                break

            history[epoch] = logs
            self._epoch += 1

        self.callbacks.on_train_end(logs)

        return history
    
    def evaluate(
        self, 
        dataProvider: DataProvider, 
        initial_epoch: int = 1,
        callbacks: typing.List[Callback] = [],
        ) -> dict:
        """ Evaluate model on validation data

        Args:
            dataProvider (DataProvider): data provider for validation data

        Returns:
            dict: validation results
        """
        self._epoch = initial_epoch
        self.callbacks = CallbacksHandler(self, callbacks)
        self.model.eval()
        logs = self.test(dataProvider)
        return logs


import tensorflow as tf


class MaskedLoss(tf.keras.losses.Loss):
    """ Masked loss function for Transformer.

    Args:
        mask_value (int, optional): Mask value. Defaults to 0.
        reduction (str, optional): Reduction method. Defaults to 'none'.
    """
    def __init__(self, mask_value: int=0, reduction: str='none') -> None:
        super(MaskedLoss, self).__init__()
        self.mask_value = mask_value
        self.reduction = reduction
        self.loss_object = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True, reduction=reduction)

    def __call__(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight=None) -> tf.Tensor:
        """ Calculate masked loss.
        
        Args:
            y_true (tf.Tensor): True labels.
            y_pred (tf.Tensor): Predicted labels.

        Returns:
            tf.Tensor: Masked loss.
        """
        mask = y_true != self.mask_value
        loss = self.loss_object(y_true, y_pred)

        mask = tf.cast(mask, dtype=loss.dtype)
        loss *= mask

        loss = tf.reduce_sum(loss) / tf.reduce_sum(mask)
        return loss


class MaskedAccuracy(tf.keras.metrics.Metric):
    """ Masked accuracy metric for Transformer.

    Args:
        mask_value (int, optional): Mask value. Defaults to 0.
        name (str, optional): Name of the metric. Defaults to 'masked_accuracy'.
    """
    def __init__(self, mask_value: int=0, name: str='masked_accuracy') -> None:
        super(MaskedAccuracy, self).__init__(name=name)
        self.mask_value = mask_value
        self.total = self.add_weight(name='total', initializer='zeros')
        self.count = self.add_weight(name='count', initializer='zeros')

    @tf.function
    def update_state(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight=None):
        """ Update state of the metric.

        Args:
            y_true (tf.Tensor): True labels.
            y_pred (tf.Tensor): Predicted labels.
        """
        pred = tf.argmax(y_pred, axis=2)
        label = tf.cast(y_true, pred.dtype)
        match = label == pred

        mask = label != self.mask_value

        match = match & mask

        match = tf.cast(match, dtype=tf.float32)
        mask = tf.cast(mask, dtype=tf.float32)
        match = tf.reduce_sum(match)
        mask = tf.reduce_sum(mask)

        self.total.assign_add(match)
        self.count.assign_add(mask)

    def result(self) -> tf.Tensor:
        """ Calculate masked accuracy.

        Returns:
            tf.Tensor: Masked accuracy.
        """
        return self.total / self.count
    

class CERMetric(tf.keras.metrics.Metric):
    """A custom TensorFlow metric to compute the Character Error Rate (CER).
    
    Args:
        vocabulary: A string of the vocabulary used to encode the labels.
        name: (Optional) string name of the metric instance.
        **kwargs: Additional keyword arguments.
    """
    def __init__(self, end_token, padding_token: int=0, name="CER", **kwargs):
        # Initialize the base Metric class
        super(CERMetric, self).__init__(name=name, **kwargs)
        
        # Initialize variables to keep track of the cumulative character/word error rates and counter
        self.cer_accumulator = tf.Variable(0.0, name="cer_accumulator", dtype=tf.float32)
        self.batch_counter = tf.Variable(0, name="batch_counter", dtype=tf.int32)
        
        self.padding_token = padding_token
        self.end_token = end_token

    def get_cer(self, pred, y_true, padding=-1):
        """ Calculates the character error rate (CER) between the predicted labels and true labels for a batch of input data.

        Args:
            pred(tf.Tensor): The predicted labels, with dtype=tf.int32, usually output from tf.keras.backend.ctc_decode
            y_true (tf.Tensor): The true labels, with dtype=tf.int32
            padding (int, optional): The padding token when converting to sparse tensor. Defaults to -1.

        Returns:
            tf.Tensor: The CER between the predicted labels and true labels
        """
        # find index where end token is
        equal = tf.equal(pred, self.end_token)
        equal_int = tf.cast(equal, tf.int64)
        end_token_index = tf.argmax(equal_int, axis=1)

        # mask out everything after end token
        new_range = tf.range(tf.shape(pred)[1], dtype=tf.int64)
        range_matrix = tf.tile(new_range[None, :], [tf.shape(pred)[0], 1])

        mask = range_matrix <= tf.expand_dims(end_token_index, axis=1)
        masked_pred = tf.where(mask, pred, padding)

        # Convert the valid predicted labels tensor to a sparse tensor
        sparse_pred = tf.RaggedTensor.from_tensor(masked_pred, padding=padding).to_sparse()

        # Convert the valid true labels tensor to a sparse tensor
        sparse_true = tf.RaggedTensor.from_tensor(y_true, padding=padding).to_sparse()

        # Calculate the normalized edit distance between the sparse predicted labels tensor and sparse true labels tensor
        distance = tf.edit_distance(sparse_pred, sparse_true, normalize=True)

        return distance

    # @tf.function
    def update_state(self, y_true, y_pred, sample_weight=None):
        """Updates the state variables of the metric.

        Args:
            y_true: A tensor of true labels with shape (batch_size, sequence_length).
            y_pred: A tensor of predicted labels with shape (batch_size, sequence_length, num_classes).
            sample_weight: (Optional) a tensor of weights with shape (batch_size, sequence_length).
        """
        pred = tf.argmax(y_pred, axis=2)

        # Calculate the normalized edit distance between the predicted labels and true labels tensors
        distance = self.get_cer(pred, y_true, self.padding_token)

        # Add the sum of the distance tensor to the cer_accumulator variable
        self.cer_accumulator.assign_add(tf.reduce_sum(distance))
        
        # Increment the batch_counter by the batch size
        self.batch_counter.assign_add(len(y_true))

    def result(self):
        """ Computes and returns the metric result.

        Returns:
            A TensorFlow float representing the CER (character error rate).
        """
        return tf.math.divide_no_nan(self.cer_accumulator, tf.cast(self.batch_counter, tf.float32))


import os
import time
import typing
import numpy as np
import onnxruntime as ort
from collections import deque

class FpsWrapper:
    """ Decorator to calculate the frames per second of a function
    """
    def __init__(self, func: typing.Callable):
        self.func = func
        self.fps_list = deque([], maxlen=100)

    def __call__(self, *args, **kwargs):
        start = time.time()
        results = self.func(self.instance, *args, **kwargs)
        self.fps_list.append(1 / (time.time() - start))
        self.instance.fps = np.mean(self.fps_list)
        return results

    def __get__(self, instance, owner):
        self.instance = instance
        return self.__call__.__get__(instance, owner)


class OnnxInferenceModel:
    """ Base class for all inference models that use onnxruntime 

    Attributes:
        model_path (str, optional): Path to the model folder. Defaults to "".
        force_cpu (bool, optional): Force the model to run on CPU or GPU. Defaults to GPU.
        default_model_name (str, optional): Default model name. Defaults to "model.onnx".
    """
    def __init__(
        self, 
        model_path: str = "",
        force_cpu: bool = False,
        default_model_name: str = "model.onnx",
        *args, **kwargs
        ):
        self.model_path = model_path.replace("\\", "/")
        self.force_cpu = force_cpu
        self.default_model_name = default_model_name

        # check if model path is a directory with os path
        if os.path.isdir(self.model_path):
            self.model_path = os.path.join(self.model_path, self.default_model_name)

        if not os.path.exists(self.model_path):
            raise Exception(f"Model path ({self.model_path}) does not exist")

        providers = ["CUDAExecutionProvider", "CPUExecutionProvider"] if ort.get_device() == "GPU" and not force_cpu else ["CPUExecutionProvider"]

        self.model = ort.InferenceSession(self.model_path, providers=providers)

        self.metadata = {}
        if self.model.get_modelmeta().custom_metadata_map:
            # add metadata to self object
            for key, value in self.model.get_modelmeta().custom_metadata_map.items():
                try:
                    new_value = eval(value) # in case the value is a list or dict
                except:
                    new_value = value
                self.metadata[key] = new_value
                
        # Update providers priority to only CPUExecutionProvider
        if self.force_cpu:
            self.model.set_providers(["CPUExecutionProvider"])

        self.input_shapes = [meta.shape for meta in self.model.get_inputs()]
        self.input_names = [meta.name for meta in self.model._inputs_meta]
        self.output_names = [meta.name for meta in self.model._outputs_meta]

    def predict(self, data: np.ndarray, *args, **kwargs):
        raise NotImplementedError

    @FpsWrapper
    def __call__(self, data: np.ndarray):
        results = self.predict(data)
        return results


import os
import typing
from pathlib import Path
import xml.etree.ElementTree as ET
from mltu.annotations.detections import Detections, Detection, BboxType

class VOCAnnotationReader:
    """Reads annotations from VOC format
    """
    def __init__(self, labels: dict, images_path: str=None):
        self.labels = labels
        self.images_path = images_path

    @staticmethod
    def readFromVOC(voc_annotation_path: str, labels: dict, images_path: str=None) -> Detections:
        annotation_path = Path(voc_annotation_path)
        tree = ET.parse(voc_annotation_path)
        root = tree.getroot()

        annotation_dict = {}

        # Iterate through child elements
        for child in root:
            if child.tag == 'object':
                obj_dict = {}
                for obj_child in child:
                    if obj_child.tag == 'bndbox':
                        bbox_dict = {}
                        for bbox_child in obj_child:
                            bbox_dict[bbox_child.tag] = int(bbox_child.text)
                        obj_dict[obj_child.tag] = bbox_dict
                    else:
                        obj_dict[obj_child.tag] = obj_child.text
                if 'objects' not in annotation_dict:
                    annotation_dict['objects'] = []
                annotation_dict['objects'].append(obj_dict)
            elif child.tag == 'size':
                size_dict = {}
                for size_child in child:
                    size_dict[size_child.tag] = int(size_child.text)
                annotation_dict['size'] = size_dict
            else:
                annotation_dict[child.tag] = child.text

        # Get the image path if not provided
        if images_path is None:
            images_path = annotation_path.parent.parent / annotation_dict["folder"]

        image_path = os.path.join(images_path, annotation_dict['filename'])
        dets = []
        for obj in annotation_dict['objects']:
            if obj['name'] not in labels.values():
                print(f"Label {obj['name']} not found in labels")
                continue

            dets.append(Detection(
                bbox=[obj['bndbox']['xmin'], obj['bndbox']['ymin'], obj['bndbox']['xmax'], obj['bndbox']['ymax']],
                label=obj['name'],
                bbox_type=BboxType.XYXY,
                confidence=1,
                image_path=image_path,
                width=annotation_dict['size']['width'],
                height=annotation_dict['size']['height'],
                relative=False
                ))
        
        detections = Detections(
            labels=labels,
            width=annotation_dict['size']['width'],
            height=annotation_dict['size']['height'],
            image_path=image_path,
            detections=dets
        )

        return detections
    
    def __call__(self, image: typing.Any, annotation: str) -> typing.Tuple[typing.Any, Detections]:
        detections = self.readFromVOC(annotation, self.labels, self.images_path)
        if image is None:
            image = detections.image_path
        return image, detections


import cv2
import numpy as np
from mltu.inferenceModel import FpsWrapper

class BaseDetector:
    """Base class for the detectors in the YOLO family"""
    @staticmethod
    def preprocess(image: np.ndarray, height: int, width: int):
        # Convert the image color space from BGR to RGB
        img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # Resize the image to match the input shape
        img = cv2.resize(img, (width, height))

        # Normalize the image data by dividing it by 255.0
        image_data = np.array(img) / 255.0

        # Transpose the image to have the channel dimension as the first dimension
        image_data = np.transpose(image_data, (2, 0, 1))  # Channel first

        # Expand the dimensions of the image data to match the expected input shape
        image_data = np.expand_dims(image_data, axis=0).astype(np.float32)

        return image_data

    @staticmethod
    def postprocess(outputs: np.ndarray, x_factor: float, y_factor: float, confidence_threshold: float=0.5, iou_threshold: float=0.5):
        # Transpose and squeeze the output to match the expected shape
        outputs = np.transpose(np.squeeze(outputs))

        # Extract all classes confidence scores
        conf_scores = np.amax(outputs[:, 4:], axis=1)

        # Get the data index of the detections with scores above the confidence threshold
        indexes = np.where(conf_scores >= confidence_threshold)[0]

        # Extract the confidence scores of the detections
        scores = conf_scores[indexes]

        # Extract the class IDs of the detections
        class_ids = np.argmax(outputs[indexes, 4:], axis=1)

        # Extract the bounding box coordinates from the outputs and transform them to the original image space
        boxes = outputs[indexes, :4] * np.array([x_factor, y_factor, x_factor, y_factor])

        # Apply non-maximum suppression to filter out overlapping bounding boxes
        indices = cv2.dnn.NMSBoxes(boxes, scores, confidence_threshold, iou_threshold)

        # Iterate over the selected indices after non-maximum suppression
        return boxes[indices], scores[indices], class_ids[indices]

    def predict(self, image: np.ndarray, **kwargs) -> np.ndarray:
        ...
        
    @FpsWrapper
    def __call__(self, image: np.ndarray):
        results = self.predict(image)
        return results